import os
from shutil import copyfile
from contextlib import ExitStack
from pathlib import Path
from collections.abc import Iterator
from subprocess import run, call
from inspect import cleandoc

from .config import Config


HOME = Path("~").expanduser()
if (HOME / ".config").exists():
    APP_PATH = HOME / ".config" / "zettel_merken"
else:
    APP_PATH = HOME / "zettel_merken"

if not APP_PATH.exists():
    os.mkdir(APP_PATH)


def get_notes_list(config: Config) -> Iterator[Path]:
    """Get a list of notes from a list of directories"""

    for dir in config.NOTE_DIRS:
        for path, dirs, files in os.walk(dir, topdown=True):

            # Exclude hidden and user ignored directories
            dirs[:] = (
                d for d in dirs if d not in config.IGNORE_DIRS and not d.startswith(".")
            )

            for file in files:
                if (
                    Path(path, file).suffix in config.INCLUDE_EXT
                    and file.casefold() not in config.IGNORE_FILES
                ):
                    yield Path(path + os.sep + file).resolve()


def get_app_path() -> Path:
    """Returns ~/.config/zettel_merken or ~/zettel_merken if .config not found"""
    return APP_PATH


def get_mail_content(notes: list[Path]) -> str:
    with ExitStack() as stack:
        files = [stack.enter_context(open(note)) for note in notes]
        return "\n\n".join(file.read() + "\n\n" for file in files)


def add_systemd_units():
    if os.name != "posix":
        print("This script for only systemd users!")
        exit()

    CONF_DIR = Path("~/.config/systemd/user").expanduser().resolve()
    TIMER_UNIT = CONF_DIR / "zettel_merken.timer"
    SERVICE_UNIT = CONF_DIR / "zettel_merken.service"

    if not CONF_DIR.exists():
        os.mkdir(CONF_DIR)

    with open(TIMER_UNIT, "w") as timer:
        timer.write(
            """\
            [Unit]
            Description=Zettel Merken Daily Review Timer
            After=network-online.target

            [Timer]
            Persistent=true
            OnCalendar=Daily
            OnBootSec=300

            [Install]
            WantedBy=timers.target\
            """
        )

    with open(SERVICE_UNIT, "w") as service:
        service.write(
            """\
            [Unit]
            Description=Zettel Merken Daily Review Service

            [Service]
            Type=simple
            ExecStart=/usr/bin/python -m zettelmerken

            [Install]
            WantedBy=default.target\
            """
        )

    run("systemctl --user daemon-reload", shell=True)
    run("systemctl --user enable --now zettel_merken.timer", shell=True)


def copy_config_file():
    if not (APP_PATH / "config.json").exists():
        copyfile(
            src=Path("./extras/config.test.json").resolve(),
            dst=APP_PATH / "config.json",
            follow_symlinks=True,
        )


def open_config():
    default_editor = "/usr/bin/vi"  # backup, if not defined in environment vars
    path = get_app_path() / "config.json"
    editor = os.environ.get("EDITOR", default_editor)
    call([editor, path])


def show_help():
    help_str = cleandoc(
        """
        USAGE: python -m zettelmerken [OPTION]

        OPTOINS:
            --help    Show this help
            --config  Create and open config.json
            --init    Init systemd units
            --remove  Remove database and systemd units
        """
    )
    print(help_str)


def remove_database():
    if APP_PATH.exists():
        (APP_PATH / "zettel_merken.db").unlink(missing_ok=True)


def remove_systemd_units():
    if os.name != "posix":
        print("This script for only systemd users!")
        exit()

    CONF_DIR = Path("~/.config/systemd/user").expanduser().resolve()
    run("systemctl --user disable --now zettel_merken.timer", shell=True)
    run("systemctl --user disable --now zettel_merken.service", shell=True)
    (CONF_DIR / "zettel_merken.timer").unlink(missing_ok=True)
    (CONF_DIR / "zettel_merken.service").unlink(missing_ok=True)



import smtplib
import ssl

from email.message import EmailMessage

from .config import Config


def send_mail(mail_content: str, config: Config) -> None:
    """Send an email containing contents of notes list scheduled for this run"""

    with smtplib.SMTP_SSL(
        config.EMAIL.HOST, config.EMAIL.PORT, context=ssl.create_default_context()
    ) as server:
        server.login(config.EMAIL.USER, config.EMAIL.PASS)

        msg = EmailMessage()
        msg["Subject"] = "Zettel Merken Daily Review"
        msg["From"] = config.EMAIL.USER
        msg["To"] = config.RECEIVERS
        msg.set_content(mail_content)

        server.send_message(msg)



import os
import json
import sqlite3

from datetime import date, timedelta
from pathlib import Path

from .config import Config


class ScheduleNotFound(Exception):
    pass


class ScheduleExhausted(Exception):
    pass


class Db:
    def __init__(self, db_path: Path, config: Config) -> None:
        self.config = config
        self.DB_PATH = db_path

    def create_schema(self) -> None:
        """Create database schema for zettel-merken note schedules"""

        with sqlite3.connect(self.DB_PATH) as cx:
            cu = cx.cursor()

            cu.execute(
                """CREATE TABLE IF NOT EXISTS note_schedule (
                    id INTEGER PRIMARY KEY ASC,
                    note TEXT NOT NULL UNIQUE,
                    stats JSON NOT NULL,
                    schedule JSON,
                    sent JSON
                )"""
            )

            cu.execute(
                "CREATE UNIQUE INDEX IF NOT EXISTS idx_note ON note_schedule (note)"
            )

    def create_note_schedule(self, note: Path, schedule=[]) -> None:
        """Create a schedule for a given note"""

        if not schedule:
            schedule = self.config.SCHEDULE_DAYS

        stats = {"mtime": os.stat(note).st_mtime}
        schedule = sorted(date.toordinal(date.today() + timedelta(s)) for s in schedule)
        sent = []

        with sqlite3.connect(self.DB_PATH) as cx:
            cx.cursor().execute(
                "INSERT INTO note_schedule (note, stats, schedule, sent) VALUES (?, ?, ?, ?)",
                (
                    str(note),
                    json.dumps(stats),
                    json.dumps(schedule),
                    json.dumps(sent),
                ),
            )

    def is_note_scheduled(self, note: Path):
        """Check if a note is scheduled for this run"""

        with sqlite3.connect(self.DB_PATH) as cx:
            data = (
                cx.cursor()
                .execute(
                    "SELECT schedule FROM note_schedule WHERE note = (?)",
                    (str(note),),
                )
                .fetchone()
            )

            if not data:
                raise ScheduleNotFound

            schedule = json.loads(data[0])

            if not schedule:
                raise ScheduleExhausted

            today = date.toordinal(date.today())
            for day in schedule:
                if day <= today:
                    return True

            return False

    def update_note_schedule(self, notes: list[Path]) -> None:
        """If a note was mailed, update in database"""

        notes_list = [str(note) for note in notes]

        with sqlite3.connect(self.DB_PATH) as cx:
            cu = cx.cursor()

            notes_to_update = cu.execute(
                "SELECT id, schedule, sent FROM note_schedule WHERE note IN"
                f" ({', '.join('?' * len(notes_list))})",
                notes_list,
            ).fetchall()

            today = date.toordinal(date.today())

            for id, schedule, sent in notes_to_update:
                schedule = json.loads(schedule)
                schedule = schedule[1:]  # Remove first element
                sent = json.loads(sent)
                sent.append(today)  # Append today's date
                cu.execute(
                    "UPDATE note_schedule SET schedule=(?), sent=(?) WHERE id=(?)",
                    (json.dumps(schedule), json.dumps(sent), id),
                )
